@@ -167,19 +167,22 @@ from six.moves import input
 
 
 from swift.common.utils import Timestamp, get_logger, ShardRange, readconf, \
-    ShardRangeList
+    ShardRangeList, non_negative_int, config_positive_int_value
 from swift.container.backend import ContainerBroker, UNSHARDED
 from swift.container.sharder import make_shard_ranges, sharding_enabled, \
     CleavingContext, process_compactible_shard_sequences, \
     find_compactible_shard_sequences, find_overlapping_ranges, \
     find_paths, rank_paths, finalize_shrinking, DEFAULT_SHARDER_CONF, \
-    ContainerSharderConf
+    ContainerSharderConf, find_paths_with_gaps, combine_shard_ranges, \
+    update_own_shard_range_stats
 
 EXIT_SUCCESS = 0
 EXIT_ERROR = 1
 EXIT_INVALID_ARGS = 2  # consistent with argparse exit code for invalid args
 EXIT_USER_QUIT = 3
 
+MIN_SHARD_RANGE_AGE_FOR_REPAIR = 4 * 3600
+
 # Some CLI options derive their default values from DEFAULT_SHARDER_CONF if
 # they have not been set. It is therefore important that the CLI parser
 # provides None as a default so that we can detect that no value was set on the
@@ -206,14 +209,36 @@ class InvalidSolutionException(ManageShardRangesException):
         self.overlapping_donors = overlapping_donors
 
 
+def wrap_for_argparse(func, msg=None):
+    """
+    Wrap the given ``func`` to catch any ``ValueError`` and raise an
+    ``argparse.ArgumentTypeError`` instead.
+
+    :param func: a function.
+    :param msg: an optional message to use with any exception that is used; if
+        not given then the string representation of the ValueError will be
+        used.
+    :return: a function wrapper.
+    """
+    def wrapped_func(*args, **kwargs):
+        try:
+            return func(*args, **kwargs)
+        except ValueError as err:
+            raise argparse.ArgumentTypeError(str(err) if msg is None else msg)
+    return wrapped_func
+
+
 def _proceed(args):
     if args.dry_run:
         choice = 'no'
     elif args.yes:
         choice = 'yes'
     else:
-        choice = input('Do you want to apply these changes to the container '
-                       'DB? [yes/N]')
+        try:
+            choice = input('Do you want to apply these changes to the '
+                           'container DB? [yes/N]')
+        except (EOFError, KeyboardInterrupt):
+            choice = 'no'
     if choice != 'yes':
         print('No changes applied')
 
@@ -225,8 +250,8 @@ def _print_shard_range(sr, level=0):
     print(indent + '%r' % sr.name)
     print(indent + '  objects: %9d, tombstones: %9d, lower: %r'
           % (sr.object_count, sr.tombstones, sr.lower_str))
-    print(indent + '    state: %9s,                        upper: %r'
-          % (sr.state_text, sr.upper_str))
+    print(indent + '    state: %9s, deleted: %d             upper: %r'
+          % (sr.state_text, sr.deleted, sr.upper_str))
 
 
 @contextmanager
@@ -356,6 +381,9 @@ def db_info(broker, args):
            if own_sr else None))
     db_state = broker.get_db_state()
     print('db_state = %s' % db_state)
+    info = broker.get_info()
+    print('object_count = %d' % info['object_count'])
+    print('bytes_used = %d' % info['bytes_used'])
     if db_state == 'sharding':
         print('Retiring db id: %s' % broker.get_brokers()[0].get_info()['id'])
         print('Cleaving context: %s' %
@@ -383,9 +411,13 @@ def delete_shard_ranges(broker, args):
             print('  - %d existing shard ranges have started sharding' %
                   [sr.state != ShardRange.FOUND
                    for sr in shard_ranges].count(True))
-        choice = input('Do you want to show the existing ranges [s], '
-                       'delete the existing ranges [yes] '
-                       'or quit without deleting [q]? ')
+        try:
+            choice = input('Do you want to show the existing ranges [s], '
+                           'delete the existing ranges [yes] '
+                           'or quit without deleting [q]? ')
+        except (EOFError, KeyboardInterrupt):
+            choice = 'q'
+
         if choice == 's':
             show_shard_ranges(broker, args)
             continue
@@ -406,6 +438,40 @@ def delete_shard_ranges(broker, args):
     return EXIT_SUCCESS
 
 
+def merge_shard_ranges(broker, args):
+    _check_own_shard_range(broker, args)
+    shard_data = _load_and_validate_shard_data(args, require_index=False)
+    new_shard_ranges = ShardRangeList([ShardRange.from_dict(sr)
+                                       for sr in shard_data])
+    new_shard_ranges.sort(key=ShardRange.sort_key)
+
+    # do some checks before merging...
+    existing_shard_ranges = ShardRangeList(
+        broker.get_shard_ranges(include_deleted=True))
+    outcome = combine_shard_ranges(new_shard_ranges, existing_shard_ranges)
+    if args.verbose:
+        print('This change will result in the following shard ranges in the '
+              'affected namespace:')
+        print(json.dumps([dict(sr) for sr in outcome], indent=2))
+    overlaps = find_overlapping_ranges(outcome)
+    if overlaps:
+        print('WARNING: this change will result in shard ranges overlaps!')
+    paths_with_gaps = find_paths_with_gaps(outcome)
+    gaps = [gap for start_path, gap, end_path in paths_with_gaps
+            if existing_shard_ranges.includes(gap)]
+    if gaps:
+        print('WARNING: this change will result in shard ranges gaps!')
+
+    if not _proceed(args):
+        return EXIT_USER_QUIT
+
+    with broker.updated_timeout(args.replace_timeout):
+        broker.merge_shard_ranges(new_shard_ranges)
+    print('Injected %d shard ranges.' % len(new_shard_ranges))
+    print('Run container-replicator to replicate them to other nodes.')
+    return EXIT_SUCCESS
+
+
 def _replace_shard_ranges(broker, args, shard_data, timeout=0):
     own_shard_range = _check_own_shard_range(broker, args)
     shard_ranges = make_shard_ranges(
@@ -449,6 +515,8 @@ def _enable_sharding(broker, own_shard_range, args):
     if own_shard_range.update_state(ShardRange.SHARDING):
         own_shard_range.epoch = Timestamp.now()
         own_shard_range.state_timestamp = own_shard_range.epoch
+    # initialise own_shard_range with current broker object stats...
+    update_own_shard_range_stats(broker, own_shard_range)
 
     with broker.updated_timeout(args.enable_timeout):
         broker.merge_shard_ranges([own_shard_range])
@@ -543,6 +611,57 @@ def compact_shard_ranges(broker, args):
     return EXIT_SUCCESS
 
 
+def _remove_illegal_overlapping_donors(
+        acceptor_path, overlapping_donors, args):
+    # Check parent-children relationship in overlaps between acceptors and
+    # donors, remove any overlapping parent or child shard range from donors.
+    # Note: we can use set() here, since shard range object is hashed by
+    # id and all shard ranges in overlapping_donors are unique already.
+    parent_child_donors = set()
+    for acceptor in acceptor_path:
+        parent_child_donors.update(
+            [donor for donor in overlapping_donors
+                if acceptor.is_child_of(donor) or donor.is_child_of(acceptor)])
+    if parent_child_donors:
+        overlapping_donors = ShardRangeList(
+            [sr for sr in overlapping_donors
+                if sr not in parent_child_donors])
+        print('%d donor shards ignored due to parent-child relationship '
+              'checks' % len(parent_child_donors))
+
+    # Check minimum age requirement in overlaps between acceptors and donors.
+    if args.min_shard_age == 0:
+        return acceptor_path, overlapping_donors
+    ts_now = Timestamp.now()
+    # Remove overlapping donor shard ranges who were created recently within
+    # 'min_shard_age' age limit.
+    qualified_donors = ShardRangeList(
+        [sr for sr in overlapping_donors
+            if float(sr.timestamp) + args.min_shard_age < float(ts_now)])
+    young_donors = len(overlapping_donors) - len(qualified_donors)
+    if young_donors > 0:
+        print('%d overlapping donor shards ignored due to minimum age '
+              'limit' % young_donors)
+    if not qualified_donors:
+        return acceptor_path, None
+    # Remove those overlapping donors whose overlapping acceptors were created
+    # within age limit.
+    donors_with_young_overlap_acceptor = set()
+    for acceptor_sr in acceptor_path:
+        if float(acceptor_sr.timestamp) + args.min_shard_age < float(ts_now):
+            continue
+        donors_with_young_overlap_acceptor.update(
+            [sr for sr in qualified_donors if acceptor_sr.overlaps(sr)])
+    if donors_with_young_overlap_acceptor:
+        qualified_donors = ShardRangeList(
+            [sr for sr in qualified_donors
+             if sr not in donors_with_young_overlap_acceptor])
+        print('%d donor shards ignored due to existence of overlapping young '
+              'acceptors' % len(donors_with_young_overlap_acceptor))
+
+    return acceptor_path, qualified_donors
+
+
 def _find_overlapping_donors(shard_ranges, own_sr, args):
     shard_ranges = ShardRangeList(shard_ranges)
     if ShardRange.SHARDING in shard_ranges.states:
@@ -593,7 +712,80 @@ def _find_overlapping_donors(shard_ranges, own_sr, args):
             'Isolated cleaved and/or active shard ranges in donor ranges',
             acceptor_path, overlapping_donors)
 
-    return acceptor_path, overlapping_donors
+    return _remove_illegal_overlapping_donors(
+        acceptor_path, overlapping_donors, args)
+
+
+def _fix_gaps(broker, args, paths_with_gaps):
+    timestamp = Timestamp.now()
+    solutions = []
+    print('Found %d gaps:' % len(paths_with_gaps))
+    for start_path, gap_range, end_path in paths_with_gaps:
+        if end_path[0].state == ShardRange.ACTIVE:
+            expanding_range = end_path[0]
+            solutions.append((gap_range, expanding_range))
+        elif start_path[-1].state == ShardRange.ACTIVE:
+            expanding_range = start_path[-1]
+            solutions.append((gap_range, expanding_range))
+        else:
+            expanding_range = None
+        print('  gap: %r - %r'
+              % (gap_range.lower, gap_range.upper))
+        print('    apparent gap contents:')
+        for sr in broker.get_shard_ranges(marker=gap_range.lower,
+                                          end_marker=gap_range.upper,
+                                          include_deleted=True):
+            _print_shard_range(sr, 3)
+        if expanding_range:
+            print('    gap can be fixed by expanding neighbor range:')
+            _print_shard_range(expanding_range, 3)
+        else:
+            print('Warning: cannot fix gap: non-ACTIVE neighbors')
+
+    if args.max_expanding >= 0:
+        solutions = solutions[:args.max_expanding]
+
+    # it's possible that an expanding range is used twice, expanding both down
+    # and up; if so, we only want one copy of it in our merged shard ranges
+    expanding_ranges = {}
+    for gap_range, expanding_range in solutions:
+        expanding_range.expand([gap_range])
+        expanding_range.timestamp = timestamp
+        expanding_ranges[expanding_range.name] = expanding_range
+
+    print('')
+    print('Repairs necessary to fill gaps.')
+    print('The following expanded shard range(s) will be applied to the DB:')
+    for expanding_range in sorted(expanding_ranges.values(),
+                                  key=lambda s: s.lower):
+        _print_shard_range(expanding_range, 2)
+    print('')
+    print(
+        'It is recommended that no other concurrent changes are made to the \n'
+        'shard ranges while fixing gaps. If necessary, abort this change \n'
+        'and stop any auto-sharding processes before repeating this command.'
+    )
+    print('')
+
+    if not _proceed(args):
+        return EXIT_USER_QUIT
+
+    broker.merge_shard_ranges(list(expanding_ranges.values()))
+    print('Run container-replicator to replicate the changes to other nodes.')
+    print('Run container-sharder on all nodes to fill gaps.')
+    return EXIT_SUCCESS
+
+
+def repair_gaps(broker, args):
+    shard_ranges = broker.get_shard_ranges()
+    paths_with_gaps = find_paths_with_gaps(shard_ranges)
+    if paths_with_gaps:
+        return _fix_gaps(broker, args, paths_with_gaps)
+    else:
+        print('Found one complete sequence of %d shard ranges with no gaps.'
+              % len(shard_ranges))
+        print('No repairs necessary.')
+        return EXIT_SUCCESS
 
 
 def print_repair_solution(acceptor_path, overlapping_donors):
@@ -647,12 +839,7 @@ def find_repair_solution(shard_ranges, own_sr, args):
     return acceptor_path, overlapping_donors
 
 
-def repair_shard_ranges(broker, args):
-    if not broker.is_root_container():
-        print('WARNING: Shard containers cannot be repaired.')
-        print('This command should be used on a root container.')
-        return EXIT_ERROR
-
+def repair_overlaps(broker, args):
     shard_ranges = broker.get_shard_ranges()
     if not shard_ranges:
         print('No shards found, nothing to do.')
@@ -682,6 +869,17 @@ def repair_shard_ranges(broker, args):
     return EXIT_SUCCESS
 
 
+def repair_shard_ranges(broker, args):
+    if not broker.is_root_container():
+        print('WARNING: Shard containers cannot be repaired.')
+        print('This command should be used on a root container.')
+        return EXIT_ERROR
+    if args.gaps:
+        return repair_gaps(broker, args)
+    else:
+        return repair_overlaps(broker, args)
+
+
 def analyze_shard_ranges(args):
     shard_data = _load_and_validate_shard_data(args, require_index=False)
     for data in shard_data:
@@ -697,13 +895,6 @@ def analyze_shard_ranges(args):
     return EXIT_SUCCESS
 
 
-def _positive_int(arg):
-    val = int(arg)
-    if val <= 0:
-        raise argparse.ArgumentTypeError('must be > 0')
-    return val
-
-
 def _add_find_args(parser):
     parser.add_argument(
         'rows_per_shard', nargs='?', type=int, default=USE_SHARDER_DEFAULT,
@@ -712,7 +903,8 @@ def _add_find_args(parser):
         'given in a conf file specified with --config, otherwise %s.'
         % DEFAULT_SHARDER_CONF['rows_per_shard'])
     parser.add_argument(
-        '--minimum-shard-size', type=_positive_int,
+        '--minimum-shard-size',
+        type=wrap_for_argparse(config_positive_int_value, 'must be > 0'),
         default=USE_SHARDER_DEFAULT,
         help='Minimum size of the final shard range. If this is greater than '
         'one then the final shard range may be extended to more than '
@@ -720,13 +912,17 @@ def _add_find_args(parser):
         'than minimum-shard-size rows.')
 
 
-def _add_replace_args(parser):
+def _add_account_prefix_arg(parser):
     parser.add_argument(
         '--shards_account_prefix', metavar='shards_account_prefix', type=str,
         required=False, default='.shards_',
         help="Prefix for shards account. The default is '.shards_'. This "
              "should only be changed if the auto_create_account_prefix option "
              "has been similarly changed in swift.conf.")
+
+
+def _add_replace_args(parser):
+    _add_account_prefix_arg(parser)
     parser.add_argument(
         '--replace-timeout', type=int, default=600,
         help='Minimum DB timeout to use when replacing shard ranges.')
@@ -756,6 +952,15 @@ def _add_prompt_args(parser):
              'Cannot be used with --yes option.')
 
 
+def _add_max_expanding_arg(parser):
+    parser.add_argument(
+        '--max-expanding', nargs='?',
+        type=wrap_for_argparse(config_positive_int_value, 'must be > 0'),
+        default=USE_SHARDER_DEFAULT,
+        help='Maximum number of shards that should be '
+        'expanded. Defaults to unlimited.')
+
+
 def _make_parser():
     parser = argparse.ArgumentParser(description='Manage shard ranges')
     parser.add_argument('path_to_file',
@@ -815,6 +1020,22 @@ def _make_parser():
         'info', help='Print container db info')
     info_parser.set_defaults(func=db_info)
 
+    # merge
+    merge_parser = subparsers.add_parser(
+        'merge',
+        help='Merge shard range(s) from file with existing shard ranges. This '
+             'subcommand should only be used if you are confident that you '
+             'know what you are doing. Shard ranges should not typically be '
+             'modified in this way.')
+    merge_parser.add_argument('input', metavar='input_file',
+                              type=str, help='Name of file')
+    merge_parser.add_argument(
+        '--replace-timeout', type=int, default=600,
+        help='Minimum DB timeout to use when merging shard ranges.')
+    _add_account_prefix_arg(merge_parser)
+    _add_prompt_args(merge_parser)
+    merge_parser.set_defaults(func=merge_shard_ranges)
+
     # replace
     replace_parser = subparsers.add_parser(
         'replace',
@@ -850,37 +1071,36 @@ def _make_parser():
              'of rows. This command only works on root containers.')
     _add_prompt_args(compact_parser)
     compact_parser.add_argument(
-        '--shrink-threshold', nargs='?', type=_positive_int,
+        '--shrink-threshold', nargs='?',
+        type=wrap_for_argparse(config_positive_int_value, 'must be > 0'),
         default=USE_SHARDER_DEFAULT,
         help='The number of rows below which a shard can qualify for '
-             'shrinking. '
-             'Defaults to %d' % DEFAULT_SHARDER_CONF['shrink_threshold'])
+        'shrinking. '
+        'Defaults to %d' % DEFAULT_SHARDER_CONF['shrink_threshold'])
     compact_parser.add_argument(
-        '--expansion-limit', nargs='?', type=_positive_int,
+        '--expansion-limit', nargs='?',
+        type=wrap_for_argparse(config_positive_int_value, 'must be > 0'),
         default=USE_SHARDER_DEFAULT,
         help='Maximum number of rows for an expanding shard to have after '
-             'compaction has completed. '
-             'Defaults to %d' % DEFAULT_SHARDER_CONF['expansion_limit'])
+        'compaction has completed. '
+        'Defaults to %d' % DEFAULT_SHARDER_CONF['expansion_limit'])
     # If just one donor shard is chosen to shrink to an acceptor then the
     # expanded acceptor will handle object listings as soon as the donor shard
     # has shrunk. If more than one donor shard are chosen to shrink to an
     # acceptor then the acceptor may not handle object listings for some donor
     # shards that have shrunk until *all* donors have shrunk, resulting in
     # temporary gap(s) in object listings where the shrunk donors are missing.
-    compact_parser.add_argument('--max-shrinking', nargs='?',
-                                type=_positive_int,
-                                default=USE_SHARDER_DEFAULT,
-                                help='Maximum number of shards that should be '
-                                     'shrunk into each expanding shard. '
-                                     'Defaults to 1. Using values greater '
-                                     'than 1 may result in temporary gaps in '
-                                     'object listings until all selected '
-                                     'shards have shrunk.')
-    compact_parser.add_argument('--max-expanding', nargs='?',
-                                type=_positive_int,
-                                default=USE_SHARDER_DEFAULT,
-                                help='Maximum number of shards that should be '
-                                     'expanded. Defaults to unlimited.')
+    compact_parser.add_argument(
+        '--max-shrinking', nargs='?',
+        type=wrap_for_argparse(config_positive_int_value, 'must be > 0'),
+        default=USE_SHARDER_DEFAULT,
+        help='Maximum number of shards that should be '
+        'shrunk into each expanding shard. '
+        'Defaults to 1. Using values greater '
+        'than 1 may result in temporary gaps in '
+        'object listings until all selected '
+        'shards have shrunk.')
+    _add_max_expanding_arg(compact_parser)
     compact_parser.set_defaults(func=compact_shard_ranges)
 
     # repair
@@ -889,6 +1109,20 @@ def _make_parser():
         help='Repair overlapping shard ranges. No action will be taken '
              'without user confirmation unless the -y option is used.')
     _add_prompt_args(repair_parser)
+    repair_parser.add_argument(
+        '--min-shard-age', nargs='?',
+        type=wrap_for_argparse(non_negative_int, 'must be >= 0'),
+        default=MIN_SHARD_RANGE_AGE_FOR_REPAIR,
+        help='Minimum age of a shard for it to be considered as an overlap '
+        'that is due for repair. Overlapping shards younger than this '
+        'age will be ignored. Value of 0 means no recent shards will be '
+        'ignored. Defaults to %d.' % MIN_SHARD_RANGE_AGE_FOR_REPAIR)
+    # TODO: maybe this should be a separate subcommand given that it needs
+    #  some extra options vs repairing overlaps?
+    repair_parser.add_argument(
+        '--gaps', action='store_true', default=False,
+        help='Repair gaps in shard ranges.')
+    _add_max_expanding_arg(repair_parser)
     repair_parser.set_defaults(func=repair_shard_ranges)
 
     # analyze
@@ -896,6 +1130,14 @@ def _make_parser():
         'analyze',
         help='Analyze shard range json data read from file. Use -v to see '
              'more detailed analysis.')
+    analyze_parser.add_argument(
+        '--min-shard-age', nargs='?',
+        type=wrap_for_argparse(non_negative_int, 'must be >= 0'),
+        default=0,
+        help='Minimum age of a shard for it to be considered as an overlap '
+        'that is due for repair. Overlapping shards younger than this '
+        'age will be ignored. Value of 0 means no recent shards will be '
+        'ignored. Defaults to 0.')
     analyze_parser.set_defaults(func=analyze_shard_ranges)
 
     return parser
